# Copyright 2019,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :class:`~zhmcclient.StorageGroupTemplate` object represents a
:term:`storage group template` object and can be used to create
:term:`storage group` objects with the properties of the template, including an
initial set of storage volumes, using the
:meth:`zhmcclient.StorageGroupTemplateManager.create` method.

Storage group template objects can be created, updated and deleted.

Storage group template objects can only be defined in CPCs that are in
DPM mode and have SE version >= 2.14.1.
"""


import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._storage_volume_template import StorageVolumeTemplateManager
from ._logging import logged_api_call
from ._utils import RC_STORAGE_TEMPLATE

__all__ = ['StorageGroupTemplateManager', 'StorageGroupTemplate']


class StorageGroupTemplateManager(BaseManager):
    """
    Manager providing access to the
    :term:`storage group templates <storage group template>` of the HMC.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable:

    * :attr:`~zhmcclient.Console.storage_group_templates` of a
      :class:`~zhmcclient.Console` object.

    HMC/SE version requirements:

    * SE version >= 2.14.1
    """

    def __init__(self, console):
        # This function should not go into the docs.
        # Parameters:
        #   console (:class:`~zhmcclient.Console`):
        #     CPC or HMC defining the scope for this manager.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'cpc-uri',
            'name',
            'type',
        ]

        super().__init__(
            resource_class=StorageGroupTemplate,
            class_name=RC_STORAGE_TEMPLATE,
            session=console.manager.session,
            parent=console,
            base_uri='/api/storage-templates',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props)
        self._console = console

    @property
    def console(self):
        """
        :class:`~zhmcclient.Console`: The Console object representing the HMC.
        """
        return self._console

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the storage group templates defined in the HMC.

        Storage group templates for which the authenticated user does not have
        object-access permission are not included.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * SE version >= 2.14.1

        Authorization requirements:

        * Object-access permission to any storage group templates to be
          included in the result.

        Parameters:

          full_properties (bool):
            Controls that the full set of resource properties for each returned
            storage group template is being retrieved, vs. only the following
            short set: "object-uri", "cpc-uri", "name", and "type".

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen.

        Returns:

          : A list of :class:`~zhmcclient.StorageGroupTemplate` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'storage-templates'
        list_uri = self._base_uri
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args, None)

    @logged_api_call
    def create(self, properties):
        """
        Create a storage group template.

        The new storage group will be associated with the CPC identified by the
        `cpc-uri` input property.

        HMC/SE version requirements:

        * SE version >= 2.14.1

        Authorization requirements:

        * Object-access permission to the CPC that will be associated with
          the new storage group template.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create Storage Template' in the :term:`HMC API` book.

            The 'cpc-uri' property identifies the CPC to which the new
            storage group template will be associated, and is required to be
            specified in this parameter.

        Returns:

          :class:`~zhmcclient.StorageGroupTemplate`:
            The resource object for the new storage group template.
            The object will have its 'object-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        if properties is None:
            properties = {}

        result = self.session.post(self._base_uri, body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        storage_group_template = StorageGroupTemplate(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return storage_group_template


class StorageGroupTemplate(BaseResource):
    """
    Representation of a :term:`storage group template`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.StorageGroupTemplateManager`).

    HMC/SE version requirements:

    * SE version >= 2.14.1
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.StorageGroupTemplateManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, StorageGroupTemplateManager), (
            "StorageGroupTemplate init: Expected manager type "
            f"{StorageGroupTemplateManager}, got {type(manager)}")
        super().__init__(
            manager, uri, name, properties)
        # The manager objects for child resources (with lazy initialization):
        self._storage_volume_templates = None
        self._cpc = None

    @property
    def storage_volume_templates(self):
        """
        :class:`~zhmcclient.StorageVolumeManager`: Access to the
        :term:`storage volumes <storage volume>` in this storage group.
        """
        # We do here some lazy loading.
        if not self._storage_volume_templates:
            self._storage_volume_templates = StorageVolumeTemplateManager(self)
        return self._storage_volume_templates

    @property
    def cpc(self):
        """
        :class:`~zhmcclient.Cpc`: The :term:`CPC` to which this storage group
        template is associated.

        The returned :class:`~zhmcclient.Cpc` has only a minimal set of
        properties populated.
        """
        # We do here some lazy loading.
        if not self._cpc:
            cpc_uri = self.get_property('cpc-uri')
            cpc_mgr = self.manager.console.manager.client.cpcs
            self._cpc = cpc_mgr.resource_object(cpc_uri)
        return self._cpc

    @logged_api_call
    def delete(self):
        """
        Delete this storage group template and its storage volume template
        resources on the HMC.

        Storage groups and their volumes that have been created from the
        template that is deleted, are not affected.

        HMC/SE version requirements:

        * SE version >= 2.14.1

        Authorization requirements:

        * Object-access permission to this storage group template.
        * Task permission to the "Configure Storage - System Programmer" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(uri=self.uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this storage group template.

        This includes the `storage-template-volumes` property which contains
        requests for creations, deletions and updates of
        :class:`~zhmcclient.StorageVolumeTemplate` resources of this storage
        group template.

        As an alternative to this bulk approach for managing storage volume
        templates, each :class:`~zhmcclient.StorageVolumeTemplate` resource
        can individually be created, deleted and updated using the respective
        methods on
        :attr:`~zhmcclient.StorageGroupTemplate.storage_volume_templates`.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * SE version >= 2.14.1

        Authorization requirements:

        * Object-access permission to this storage group template.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are listed for operation
            'Modify Storage Template Properties' in the :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        uri = f'{self.uri}/operations/modify'
        self.manager.session.post(uri, resource=self, body=properties)
        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)



# Copyright 2018,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Starting with the z14-ZR1 and LinuxONE Rockhopper II machine generations, the
"dpm-storage-management" :ref:`firmware feature <firmware features>` has been
introduced to support a simpler management of FCP and FICON (=ECKD) storage for
DPM mode. If machines of these generations are in DPM mode, the feature is
always enabled and cannot be disabled.

When the "dpm-storage-management" :ref:`firmware feature <firmware features>`
is enabled, :term:`storage group`
and :term:`storage volume` resources can be defined on the HMC, and can cause
fulfillment requests to be sent via email to storage administrators.
Once these requests are satisfied on the actual storage subsystem and possibly
SAN switches, the changes are automatically discovered by DPM and reflected
in the state of these resources. Thus, the allocation of actual storage volumes
on the storage subsystem is not performed by DPM.

The sending of email is optional, and if the changes are done by some
automation tool, they will also be discovered automatically. That way, the
whole process of allocating and attaching volumes can be fully automated, if
so desired.

The top level resource objects are :term:`storage groups <storage group>`.
Storage groups are defined globally at the HMC level, and are associated with a
CPC. They can only be associated with one CPC at a time. In the zhmcclient, the
:class:`~zhmcclient.StorageGroup` objects are accessible via the
:attr:`~zhmcclient.ConsoleManager.storage_groups` property.

Storage groups are a grouping mechanism for
:term:`storage volumes <storage volume>`. An FCP-type storage
group can contain only FCP type storage volumes, and a FICON-type storage
group can contain only FICON/ECKD type storage volumes.

Attachment and detachment of volumes to a partition happens at the level of
storage groups, and always applies to all volumes defined in the storage group.

The storage-related z/Architecture devices that are visible to a partition are
different for the two storage architectures: For FCP, the virtualized HBA is
visible as a device, and the storage volumes (LUNs) are not represented as
devices. For FICON, each ECKD volume is visible as a device, but the
virtualized FICON adapter port is not represented as a device. When the
"dpm-storage-management" :ref:`firmware feature <firmware features>` is enabled,
each storage-related z/Architecture device that is visible to a partition is
represented as a :term:`virtual storage resource` object. The virtual storage
resource objects are instantiated automatically when a storage group is attached
to a partition. The :term:`HBA` resource objects known from DPM mode before the
introduction of the "dpm-storage-management"
:ref:`firmware feature <firmware features>` are not exposed anymore (their
equivalent are now the :term:`virtual storage resource` objects).

Single storage volumes cannot be attached to partitions, only entire storage
groups can be. In fact, storage volume objects do not even exist outside the
scope of storage groups.
A particular storage volume can be contained in only one storage group.

Storage groups can be listed, created, deleted, and updated, and their storage
volumes can also be listed, created, deleted, and updated.

Storage groups can be attached to zero or more partitions. Attachment to
multiple partitions at the same time is possible for storage groups that are
defined to be shareable. In case of multiple attachments of a storage group, it
contains the storage volume objects only once for all attachments, but the
virtual storage resource objects are instantiated separately for each
attachment.

Storage groups can only be associated with CPCs that have the
"dpm-storage-management" :ref:`firmware feature <firmware features>` enabled.
"""


import copy
import re

from ._manager import BaseManager
from ._resource import BaseResource
from ._storage_volume import StorageVolumeManager
from ._virtual_storage_resource import VirtualStorageResourceManager
from ._logging import logged_api_call
from ._utils import append_query_parms, RC_STORAGE_GROUP

__all__ = ['StorageGroupManager', 'StorageGroup']


class StorageGroupManager(BaseManager):
    """
    Manager providing access to the :term:`storage groups <storage group>` of
    the HMC.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable:

    * :attr:`~zhmcclient.Console.storage_groups` of a
      :class:`~zhmcclient.Console` object.

    HMC/SE version requirements:

    * :ref:`firmware feature <firmware features>` "dpm-storage-management"
    """

    def __init__(self, console):
        # This function should not go into the docs.
        # Parameters:
        #   console (:class:`~zhmcclient.Console`):
        #     CPC or HMC defining the scope for this manager.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'cpc-uri',
            'name',
            'type',
            'fulfillment-state',
        ]

        super().__init__(
            resource_class=StorageGroup,
            class_name=RC_STORAGE_GROUP,
            session=console.manager.session,
            parent=console,
            base_uri='/api/storage-groups',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props)
        self._console = console

    @property
    def console(self):
        """
        :class:`~zhmcclient.Console`: The Console object representing the HMC.
        """
        return self._console

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the storage groups defined in the HMC.

        Storage groups for which the authenticated user does not have
        object-access permission are not included.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to any storage groups to be included in the
          result.

        Parameters:

          full_properties (bool):
            Controls that the full set of resource properties for each returned
            storage volume is being retrieved, vs. only the following short
            set: "object-uri", "cpc-uri", "name", "fulfillment-state", and
            "type".

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen.

        Returns:

          : A list of :class:`~zhmcclient.StorageGroup` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'storage-groups'
        list_uri = self._base_uri
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args, None)

    @logged_api_call
    def create(self, properties=None, template=None):
        """
        Create and configure a storage group either from input properties or
        from a :term:`storage group template`.

        The input properties may specify initial storage volumes for the new
        storage group via the `storage-volumes` property. Additional storage
        volumes can be added with
        :meth:`~zhmcclient.StorageGroup.update_properties`.

        A storage group template can also specify initial storage volumes for
        the new storage group. For details, see
        :class:`~zhmcclient.StorageGroupTemplate`.

        The new storage group will be associated with the CPC identified by the
        `cpc-uri` input property.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to the CPC that will be associated with
          the new storage group.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create Storage Group' in the :term:`HMC API` book.

            The 'cpc-uri' property identifies the CPC to which the new
            storage group will be associated, and is required to be specified
            in this parameter.

            The 'template-uri' property is not allowed. If you want to create
            a storage group from a :term:`storage group template`, specify the
            `template` parameter.

            The `properties` and `template` parameters are mutually exclusive.

          template (:class:`~zhmcclient.StorageGroupTemplate`):
            :term:`storage group template` defining the initial property values
            for the storage group, including its initial storage volumes.

        Returns:

          :class:`~zhmcclient.StorageGroup`:
            The resource object for the new storage group.
            The object will have its 'object-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`ValueError` - Property 'template-uri' is not permitted
          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        if properties is None:
            properties = {}

        if 'template-uri' in properties:
            raise ValueError(
                "Property 'template-uri' is not permitted - use the "
                "'template' parameter to create storage groups from "
                "templates.")

        if template is not None:
            properties['template-uri'] = template.uri

        result = self.session.post(self._base_uri, body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        storage_group = StorageGroup(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return storage_group


class StorageGroup(BaseResource):
    """
    Representation of a :term:`storage group`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.StorageGroupManager`).

    HMC/SE version requirements:

    * :ref:`firmware feature <firmware features>` "dpm-storage-management"
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.StorageGroupManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, StorageGroupManager), (
            f"StorageGroup init: Expected manager type {StorageGroupManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)
        # The manager objects for child resources (with lazy initialization):
        self._storage_volumes = None
        self._virtual_storage_resources = None
        self._cpc = None

    @property
    def storage_volumes(self):
        """
        :class:`~zhmcclient.StorageVolumeManager`: Access to the
        :term:`storage volumes <storage volume>` in this storage group.
        """
        # We do here some lazy loading.
        if not self._storage_volumes:
            self._storage_volumes = StorageVolumeManager(self)
        return self._storage_volumes

    @property
    def virtual_storage_resources(self):
        """
        :class:`~zhmcclient.VirtualStorageResourceManager`: Access to the
        :term:`virtual storage resources <Virtual Storage Resource>` in this
        storage group.
        """
        # We do here some lazy loading.
        if not self._virtual_storage_resources:
            self._virtual_storage_resources = VirtualStorageResourceManager(
                self)
        return self._virtual_storage_resources

    @property
    def cpc(self):
        """
        :class:`~zhmcclient.Cpc`: The :term:`CPC` to which this storage group
        is associated.

        The returned :class:`~zhmcclient.Cpc` has only a minimal set of
        properties populated.
        """
        # We do here some lazy loading.
        if not self._cpc:
            cpc_uri = self.get_property('cpc-uri')
            cpc_mgr = self.manager.console.manager.client.cpcs
            self._cpc = cpc_mgr.resource_object(cpc_uri)
        return self._cpc

    @logged_api_call
    def list_attached_partitions(self, name=None, status=None):
        """
        Return the partitions to which this storage group is currently
        attached, optionally filtered by partition name and status.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this storage group.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          name (:term:`string`): Filter pattern (regular expression) to limit
            returned partitions to those that have a matching name. If `None`,
            no filtering for the partition name takes place.

          status (:term:`string`): Filter string to limit returned partitions
            to those  that have a matching status. The value must be a valid
            partition status property value. If `None`, no filtering for the
            partition status takes place.

        Returns:

          List of :class:`~zhmcclient.Partition` objects representing the
          partitions to whivch this storage group is currently attached,
          with a minimal set of properties ('object-id', 'name', 'status').

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """

        query_parms = []
        if name is not None:
            append_query_parms(query_parms, 'name', name)
        if status is not None:
            append_query_parms(query_parms, 'status', status)
        query_parms_str = '&'.join(query_parms)
        if query_parms_str:
            query_parms_str = f'?{query_parms_str}'

        uri = f'{self.uri}/operations/get-partitions{query_parms_str}'

        sg_cpc = self.cpc
        part_mgr = sg_cpc.partitions

        result = self.manager.session.get(uri, resource=self)
        props_list = result['partitions']
        part_list = []
        for props in props_list:
            part = part_mgr.resource_object(props['object-uri'], props)
            part_list.append(part)
        return part_list

    @logged_api_call
    def delete(self, email_to_addresses=None, email_cc_addresses=None,
               email_insert=None):
        """
        Delete this storage group and its storage volume resources on the HMC,
        and optionally send emails to storage administrators requesting
        deletion of the storage volumes on the storage subsystem and cleanup of
        any resources related to the storage group (e.g. zones on a SAN switch,
        or host objects on a storage subsystem).

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this storage group.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          email_to_addresses (:term:`iterable` of :term:`string`): Email
            addresses of one or more storage administrator to be notified.
            If `None` or empty, no email will be sent.

          email_cc_addresses (:term:`iterable` of :term:`string`): Email
            addresses of one or more storage administrator to be copied
            on the notification email.
            If `None` or empty, nobody will be copied on the email.
            Must be `None` or empty if `email_to_addresses` is `None` or empty.

          email_insert (:term:`string`): Additional text to be inserted in the
            notification email.
            The text can include HTML formatting tags.
            If `None`, no additional text will be inserted.
            Must be `None` or empty if `email_to_addresses` is `None` or empty.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`ValueError`: Incorrect input arguments
        """

        body = {}

        if email_to_addresses:
            body['email-to-addresses'] = email_to_addresses
            if email_cc_addresses:
                body['email-cc-addresses'] = email_cc_addresses
            if email_insert:
                body['email-insert'] = email_insert
        else:
            if email_cc_addresses:
                raise ValueError(
                    "email_cc_addresses must not be specified if there is no "
                    f"email_to_addresses: {email_cc_addresses!r}")
            if email_insert:
                raise ValueError(
                    "email_insert must not be specified if there is no "
                    f"email_to_addresses: {email_insert!r}")

        self.manager.session.post(
            uri=self.uri + '/operations/delete', resource=self, body=body)

        # pylint: disable=protected-access
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this storage group.

        This includes the `storage-volumes` property which contains requests
        for creations, deletions and updates of
        :class:`~zhmcclient.StorageVolume` resources of this storage group.
        As an alternative to this bulk approach for managing storage volumes,
        each :class:`~zhmcclient.StorageVolume` resource can individually be
        created, deleted and updated using the respective methods on
        :attr:`~zhmcclient.StorageGroup.storage_volumes`.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this storage group.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are listed for operation
            'Modify Storage Group Properties' in section 'Storage Group object'
            in the :term:`HMC API` book. This includes the email addresses
            of the storage administrators.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        uri = f'{self.uri}/operations/modify'
        self.manager.session.post(uri, resource=self, body=properties)
        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)

    @logged_api_call
    def add_candidate_adapter_ports(self, ports):
        """
        Add a list of storage adapter ports to this storage group's candidate
        adapter ports list.

        This operation only applies to storage groups of type "fcp".

        These adapter ports become candidates for use as backing adapters when
        creating virtual storage resources when the storage group is attached
        to a partition. The adapter ports should have connectivity to the
        storage area network (SAN).

        Candidate adapter ports may only be added before the CPC discovers a
        working communications path, indicated by a "verified" status on at
        least one of this storage group's WWPNs. After that point, all
        adapter ports in the storage group are automatically detected and
        manually adding them is no longer possible.

        Because the CPC discovers working communications paths automatically,
        candidate adapter ports do not need to be added by the user. Any
        ports that are added, are validated by the CPC during discovery,
        and may or may not actually be used.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this storage group.
        * Object-access permission to the adapter of each specified port.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          ports (:class:`py:list`): List of :class:`~zhmcclient.Port` objects
            representing the ports to be added. All specified ports must not
            already be members of this storage group's candidate adapter ports
            list.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {
            'adapter-port-uris': [p.uri for p in ports],
        }
        self.manager.session.post(
            self.uri + '/operations/add-candidate-adapter-ports', resource=self,
            body=body)

    @logged_api_call
    def remove_candidate_adapter_ports(self, ports):
        """
        Remove a list of storage adapter ports from this storage group's
        candidate adapter ports list.

        This operation only applies to storage groups of type "fcp".

        Because the CPC discovers working communications paths automatically,
        candidate adapter ports do not need to be managed by the user. Any
        ports that are removed using this function, might actually be added
        again by the CPC dependent on the results of discovery.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this storage group.
        * Object-access permission to the adapter of each specified port.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          ports (:class:`py:list`): List of :class:`~zhmcclient.Port` objects
            representing the ports to be removed. All specified ports must
            currently be members of this storage group's candidate adapter
            ports list and must not be referenced by any of the group's virtual
            storage resources.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {
            'adapter-port-uris': [p.uri for p in ports],
        }
        self.manager.session.post(
            self.uri + '/operations/remove-candidate-adapter-ports',
            resource=self, body=body)

    @logged_api_call
    def list_candidate_adapter_ports(self, full_properties=False):
        """
        Return the current candidate storage adapter port list of this FCP
        storage group.

        This operation only applies to storage groups of type "fcp".

        The result reflects the actual list of ports used by the CPC, including
        any changes that have been made during discovery. The source for this
        information is the 'candidate-adapter-port-uris' property of the
        storage group object.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Parameters:

          full_properties (bool):
            Controls that the full set of resource properties for each returned
            candidate storage adapter port is being retrieved, vs. only the
            following short set: "element-uri", "element-id", "class",
            "parent".

            TODO: Verify short list of properties.

        Returns:

          List of :class:`~zhmcclient.Port` objects representing the
          current candidate storage adapter ports of this storage group.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        sg_cpc = self.cpc
        adapter_mgr = sg_cpc.adapters
        port_list = []
        port_uris = self.get_property('candidate-adapter-port-uris')
        if port_uris:
            for port_uri in port_uris:
                m = re.match(r'^(/api/adapters/[^/]*)/.*', port_uri)

                adapter_uri = m.group(1)
                adapter = adapter_mgr.resource_object(adapter_uri)

                port_mgr = adapter.ports
                port = port_mgr.resource_object(port_uri)
                port_list.append(port)
                if full_properties:
                    port.pull_full_properties()

        return port_list

    @logged_api_call
    def discover_fcp(
            self, force_restart=False, wait_for_completion=True,
            operation_timeout=None):
        """
        Perform Logical Unit Number (LUN) discovery for this FCP storage group.
        The corresponding HMC operation is "Start FCP Storage Discovery".

        This operation only applies to storage groups of type "fcp".

        The HMC performs the LUN discovery in an asynchronous job. This method
        supports waiting for completion of the job and thus the LUN discovery,
        or returning immediately after the HMC has started the asynchronous job.

        When the LUN discovery is completed, a connection report will have
        been created that reflects the results of the discovery. The
        connection report can be retrieved with
        :meth:`~zhmcclient.StorageGroup.get_connection_report`. It is
        recommended to retrieve the connection report and verify that it
        reports correct volumes.

        HMC/SE version requirements:

        * HMC Version >= 2.14.1 with HMC API version >= 2.37
        * SE Version >= 2.14.1

        Authorization requirements:

        * Object-access permission to this storage group.
        * Task permission to the "Configure Storage - System Programmer" task
          or to the "Configure Storage - Storage Administrator" task.

        Parameters:

          force_restart (bool):
            Indicates if there is an in-progress discovery operation for the
            specified storage group, it should be terminated and started again.
            If `False` or there is no in-progress discovery operation for the
            specified storage group, a new one is started.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the asynchronous job performing the LUN discovery.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the LUN discovery. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns nothing.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        body = {
            'force-restart': force_restart,
        }
        result = self.manager.session.post(
            self.uri + '/operations/start-fcp-storage-discovery',
            resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        return result

    @logged_api_call
    def get_connection_report(self):
        # pylint: disable=line-too-long
        """
        Get the latest connection report for this storage group.

        The corresponding HMC operation is "Get Connection Report".

        The connection report is updated when LUN discovery is performed. LUN
        discovery for FCP storage groups is performed automatically based on
        certain triggers, and regularly every 10 minutes, and can additionally
        be triggered with :meth:`~zhmcclient.StorageGroup.discover_fcp`.

        To verify that the correct volumes are reported for an FCP storage
        group, verify that the property `volumes-configuration-status` has
        a value of "correct-volumes" for all WWPNs in all FCP subsystems::

            report = storage_group.get_connection_report()
            for subsys in report['fcp-storage-subsystems']:
                for config in subsys['storage-configurations']:
                    if config['volumes-configuration-status'] != "correct-volumes":
                        # report incorrect volumes

        HMC/SE version requirements:

        * HMC Version >= 2.14.1 with HMC API version >= 2.37
        * SE Version >= 2.14.1

        Authorization requirements:

        * Object-access permission to this storage group.
        * Task permission to the "Configure Storage - System Programmer" task
          or to the "Configure Storage - Storage Administrator" task.

        Returns:

          :term:`json object`:
            A JSON object with the connection report. For details about the
            items in the JSON object, see section 'Response body contents' in
            section 'Get Connection Report' in the :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """  # noqa: E501
        result = self.manager.session.get(
            self.uri + '/operations/get-connection-report', resource=self)
        return result

    # TODO: Add support for "Fulfill FICON Storage Volumes" operation

    # TODO: Add support for "Accept Mismatched FCP Storage Volumes" operation

    # TODO: Add support for "Reject Mismatched FCP Storage Volumes" operation

    # TODO: Add support for "Get Storage Group Histories" operation

    def dump(self):
        """
        Dump this StorageGroup resource with its properties and child
        resources (recursively) as a resource definition.

        The returned resource definition has the following format::

            {
                # Resource properties:
                "properties": {...},

                # Child resources:
                "storage_volumes": [...],
            }

        Returns:

          dict: Resource definition of this resource.
        """

        # Dump the resource properties
        resource_dict = super().dump()

        # Dump the child resources
        storage_volumes = self.storage_volumes.dump()
        if storage_volumes:
            resource_dict['storage_volumes'] = storage_volumes

        return resource_dict



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for storage volume templates (on CPCs in DPM mode).

These tests do not change any existing storage volume templates, but create,
modify and delete test storage volume templates.
"""


import warnings
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    skipif_no_storage_mgmt_feature, runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic StorageVolume objects (e.g. find_by_name())
STOVOLTPL_MINIMAL_PROPS = ['element-uri', 'name']

# Properties in StorageVolume objects returned by list() without full props
STOVOLTPL_LIST_PROPS = ['element-uri', 'name', 'size', 'usage']

# Properties whose values can change between retrievals of StorageVolume objs
STOVOLTPL_VOLATILE_PROPS = []


def test_stovoltpl_find_list(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_storage_mgmt_feature(cpc)

        console = cpc.manager.client.consoles.console
        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the storage volume templates to test with
        grp_vol_tuples = []
        stogrptpl_list = console.storage_group_templates.findall(
            **{'cpc-uri': cpc.uri})
        for stogrptpl in stogrptpl_list:
            stovoltpl_list = stogrptpl.storage_volume_templates.list()
            for stovoltpl in stovoltpl_list:
                grp_vol_tuples.append((stogrptpl, stovoltpl))
        if not grp_vol_tuples:
            skip_warn(
                "No storage group templates with volumes associated to "
                f"CPC {cpc.name} managed by HMC {hd.host}")
        grp_vol_tuples = pick_test_resources(grp_vol_tuples)

        for stogrptpl, stovoltpl in grp_vol_tuples:
            print(f"Testing on CPC {cpc.name} with storage volume template "
                  f"{stovoltpl.name!r} of storage group template "
                  f"{stogrptpl.name!r}")
            runtest_find_list(
                session, stogrptpl.storage_volume_templates, stovoltpl.name,
                'name', 'size', STOVOLTPL_VOLATILE_PROPS,
                STOVOLTPL_MINIMAL_PROPS, STOVOLTPL_LIST_PROPS)


def test_stovoltpl_property(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_storage_mgmt_feature(cpc)

        console = cpc.manager.client.consoles.console
        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the storage volume templates to test with
        grp_vol_tuples = []
        stogrptpl_list = console.storage_group_templates.findall(
            **{'cpc-uri': cpc.uri})
        for stogrptpl in stogrptpl_list:
            stovoltpl_list = stogrptpl.storage_volume_templates.list()
            for stovoltpl in stovoltpl_list:
                grp_vol_tuples.append((stogrptpl, stovoltpl))
        if not grp_vol_tuples:
            skip_warn(
                "No storage group templates with volumes associated to "
                f"CPC {cpc.name} managed by HMC {hd.host}")
        grp_vol_tuples = pick_test_resources(grp_vol_tuples)

        for stogrptpl, stovoltpl in grp_vol_tuples:
            print(f"Testing on CPC {cpc.name} with storage volume template "
                  f"{stovoltpl.name!r} of storage group template "
                  f"{stogrptpl.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(stovoltpl.manager, non_list_prop)


def test_stovoltpl_crud(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a storage volume template in a storage
    group template.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_storage_mgmt_feature(cpc)

        print(f"Testing on CPC {cpc.name}")

        console = cpc.manager.client.consoles.console

        stogrptpl_name = TEST_PREFIX + ' test_stovoltpl_crud stogrptpl1'
        stovoltpl_name = 'stovoltpl1'
        stovoltpl_name_new = stovoltpl_name + ' new'

        # Ensure clean starting point
        try:
            stogrptpl = console.storage_group_templates.find(
                name=stogrptpl_name)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                "Deleting test storage group template from previous run: "
                f"{stogrptpl_name!r} on CPC {cpc.name}", UserWarning)
            stogrptpl.delete()

        stogrptpl = None
        try:

            # Create a storage group template for the volume
            stogrptpl_input_props = {
                'cpc-uri': cpc.uri,
                'name': stogrptpl_name,
                'description': 'Dummy storage group template description.',
                'type': 'fcp',
            }
            stogrptpl = console.storage_group_templates.create(
                stogrptpl_input_props)

            # Test creating a volume

            stovoltpl_input_props = {
                'name': stovoltpl_name,
                'description': 'Dummy storage volume template description.',
                'size': 100,  # MB
            }
            stovoltpl_auto_props = {
                'usage': 'data',
            }

            # The code to be tested
            stovoltpl = stogrptpl.storage_volume_templates.create(
                stovoltpl_input_props)

            for pn, exp_value in stovoltpl_input_props.items():
                assert stovoltpl.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r} of storage volume " \
                    f"template:\n{stovoltpl.properties!r}"
            stovoltpl.pull_full_properties()
            for pn, exp_value in stovoltpl_input_props.items():
                assert stovoltpl.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r} of storage volume " \
                    f"template:\n{stovoltpl.properties!r}"
            for pn, exp_value in stovoltpl_auto_props.items():
                assert pn in stovoltpl.properties, (
                    f"Automatically returned property {pn!r} is not in "
                    f"created storage volume template:\n{stovoltpl!r}")
                assert stovoltpl.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r} of storage volume " \
                    f"template:\n{stovoltpl.properties!r}"

            # Test updating a property of the storage volume template

            new_desc = "Updated storage volume template description."

            # The code to be tested
            stovoltpl.update_properties(dict(description=new_desc))

            assert stovoltpl.properties['description'] == new_desc
            stovoltpl.pull_full_properties()
            assert stovoltpl.properties['description'] == new_desc

            # Test renaming the storage volume template

            # The code to be tested
            stovoltpl.update_properties(dict(name=stovoltpl_name_new))

            assert stovoltpl.properties['name'] == stovoltpl_name_new
            stovoltpl.pull_full_properties()
            assert stovoltpl.properties['name'] == stovoltpl_name_new
            with pytest.raises(zhmcclient.NotFound):
                stogrptpl.storage_volume_templates.find(name=stovoltpl_name)

            # Test deleting the storage volume template

            # The code to be tested
            stovoltpl.delete()

            with pytest.raises(zhmcclient.NotFound):
                stogrptpl.storage_volume_templates.find(name=stovoltpl_name_new)

        finally:
            # Cleanup
            if stogrptpl:
                stogrptpl.delete()



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Public constants.

These constants are not meant to be changed by the user, they are made
available for inspection and documentation purposes only.

For technical reasons, the online documentation shows these constants in the
``zhmcclient._constants`` namespace, but they are also available in the
``zhmcclient`` namespace and should be used from there.
"""

__all__ = ['DEFAULT_CONNECT_TIMEOUT',
           'DEFAULT_CONNECT_RETRIES',
           'DEFAULT_HMC_PORT',
           'DEFAULT_READ_TIMEOUT',
           'DEFAULT_READ_RETRIES',
           'DEFAULT_STOMP_PORT',
           'DEFAULT_MAX_REDIRECTS',
           'DEFAULT_OPERATION_TIMEOUT',
           'DEFAULT_STATUS_TIMEOUT',
           'DEFAULT_NAME_URI_CACHE_TIMETOLIVE',
           'DEFAULT_STOMP_CONNECT_TIMEOUT',
           'DEFAULT_STOMP_CONNECT_RETRIES',
           'DEFAULT_STOMP_RECONNECT_SLEEP_INITIAL',
           'DEFAULT_STOMP_RECONNECT_SLEEP_INCREASE',
           'DEFAULT_STOMP_RECONNECT_SLEEP_MAX',
           'DEFAULT_STOMP_RECONNECT_SLEEP_JITTER',
           'DEFAULT_STOMP_KEEPALIVE',
           'DEFAULT_STOMP_HEARTBEAT_SEND_CYCLE',
           'DEFAULT_STOMP_HEARTBEAT_RECEIVE_CYCLE',
           'DEFAULT_STOMP_HEARTBEAT_RECEIVE_CHECK',
           'HMC_LOGGER_NAME',
           'JMS_LOGGER_NAME',
           'API_LOGGER_NAME',
           'OS_LOGGER_NAME',
           'HTML_REASON_WEB_SERVICES_DISABLED',
           'HTML_REASON_OTHER',
           'STOMP_MIN_CONNECTION_CHECK_TIME',
           'DEFAULT_WS_TIMEOUT',
           'BLANKED_OUT_STRING']


#: Default HTTP connect timeout in seconds,
#: if not specified in the ``retry_timeout_config`` init argument to
#: :class:`~zhmcclient.Session`.
DEFAULT_CONNECT_TIMEOUT = 30

#: Default number of HTTP connect retries,
#: if not specified in the ``retry_timeout_config`` init argument to
#: :class:`~zhmcclient.Session`.
DEFAULT_CONNECT_RETRIES = 3

#: Default HMC port number
DEFAULT_HMC_PORT = 6794

#: Default HTTP read timeout in seconds,
#: if not specified in the ``retry_timeout_config`` init argument to
#: :class:`~zhmcclient.Session`.
#:
#: Note: The default value for this parameter has been increased to a large
#: value in order to mitigate the behavior of the 'requests' module to
#: retry HTTP methods even if they are not idempotent (e.g. DELETE).
#: See zhmcclient `issue #249
#: <https://github.com/zhmcclient/python-zhmcclient/issues/249>`_.
DEFAULT_READ_TIMEOUT = 3600

#: Default port on which the HMC issues JMS over STOMP messages.
DEFAULT_STOMP_PORT = 61612

#: Default number of HTTP read retries,
#: if not specified in the ``retry_timeout_config`` init argument to
#: :class:`~zhmcclient.Session`.
#:
#: Note: The default value for this parameter has been set to 0 in order to
#: mitigate the behavior of the 'requests' module to retry HTTP methods even if
#: they are not idempotent (e.g. DELETE).
#: See zhmcclient `issue #249
#: <https://github.com/zhmcclient/python-zhmcclient/issues/249>`_.
DEFAULT_READ_RETRIES = 0

#: Default max. number of HTTP redirects,
#: if not specified in the ``retry_timeout_config`` init argument to
#: :class:`~zhmcclient.Session`.
DEFAULT_MAX_REDIRECTS = 30

#: Default timeout in seconds for waiting for completion of an asynchronous
#: HMC operation,
#: if not specified in the ``retry_timeout_config`` init argument to
#: :class:`~zhmcclient.Session`.
#:
#: This is used as a default value in asynchronous methods on
#: resource objects (e.g. :meth:`zhmcclient.Partition.start`), in the
#: :meth:`zhmcclient.Job.wait_for_completion` method, and in the
#: low level method :meth:`zhmcclient.Session.post`.
DEFAULT_OPERATION_TIMEOUT = 3600

#: Default timeout in seconds for waiting for completion of deferred status
#: changes for LPARs and Partitions,
#: if not specified in the ``retry_timeout_config`` init argument to
#: :class:`~zhmcclient.Session`.
DEFAULT_STATUS_TIMEOUT = 60

#: Default time to the next automatic invalidation of the Name-URI cache of
#: manager objects, in seconds since the last invalidation,
#: if not specified in the ``retry_timeout_config`` init argument to
#: :class:`~zhmcclient.Session`.
#:
#: The special value 0 means that no Name-URI cache is maintained (i.e. the
#: caching is disabled).
DEFAULT_NAME_URI_CACHE_TIMETOLIVE = 300

#: Default value for the `connect_timeout` property of the
#: :class:`~zhmcclient.StompRetryTimeoutConfig` configuration.
DEFAULT_STOMP_CONNECT_TIMEOUT = 30

#: Default value for the `connect_retries` property of the
#: :class:`~zhmcclient.StompRetryTimeoutConfig` configuration.
DEFAULT_STOMP_CONNECT_RETRIES = 3

#: Default value for the `reconnect_sleep_initial` property of the
#: :class:`~zhmcclient.StompRetryTimeoutConfig` configuration.
DEFAULT_STOMP_RECONNECT_SLEEP_INITIAL = 0.1

#: Default value for the `reconnect_sleep_increase` property of the
#: :class:`~zhmcclient.StompRetryTimeoutConfig` configuration.
DEFAULT_STOMP_RECONNECT_SLEEP_INCREASE = 0.5

#: Default value for the `reconnect_sleep_max` property of the
#: :class:`~zhmcclient.StompRetryTimeoutConfig` configuration.
DEFAULT_STOMP_RECONNECT_SLEEP_MAX = 60

#: Default value for the `reconnect_sleep_jitter` property of the
#: :class:`~zhmcclient.StompRetryTimeoutConfig` configuration.
DEFAULT_STOMP_RECONNECT_SLEEP_JITTER = 0.1

#: Default value for the `keepalive` property of the
#: :class:`~zhmcclient.StompRetryTimeoutConfig` configuration.
DEFAULT_STOMP_KEEPALIVE = True

#: Default value for the `heartbeat_send_cycle` property of the
#: :class:`~zhmcclient.StompRetryTimeoutConfig` configuration.
DEFAULT_STOMP_HEARTBEAT_SEND_CYCLE = 5.0

#: Default value for the `heartbeat_receive_cycle` property of the
#: :class:`~zhmcclient.StompRetryTimeoutConfig` configuration.
DEFAULT_STOMP_HEARTBEAT_RECEIVE_CYCLE = 5.0

#: Default value for the `heartbeat_receive_check` property of the
#: :class:`~zhmcclient.StompRetryTimeoutConfig` configuration.
DEFAULT_STOMP_HEARTBEAT_RECEIVE_CHECK = 1.0

#: Name of the Python logger that logs HMC operations.
HMC_LOGGER_NAME = 'zhmcclient.hmc'

#: Name of the Python logger that logs zhmcclient API calls made by the user.
API_LOGGER_NAME = 'zhmcclient.api'

#: Name of the Python logger that logs JMS notifications.
JMS_LOGGER_NAME = 'zhmcclient.jms'

#: Name of the Python logger that logs interactions with OS consoles.
OS_LOGGER_NAME = 'zhmcclient.os'

#: HTTP reason code: Web Services API is not enabled on the HMC.
HTML_REASON_WEB_SERVICES_DISABLED = 900

#: HTTP reason code: Other HTML-formatted error response. Note that over time,
#: there may be more specific reason codes introduced for such situations.
HTML_REASON_OTHER = 999

#: Minimum time between checks for STOMP connection loss.
STOMP_MIN_CONNECTION_CHECK_TIME = 5.0

#: Default WebSocket connect and receive timeout in seconds, for interacting
#: with the :class:`zhmcclient.OSConsole` class.
DEFAULT_WS_TIMEOUT = 5

#: Replacement string for blanked out sensitive values in log entries, such as
#: passwords or session tokens.
BLANKED_OUT_STRING = '********'



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _utils module.
"""


import os
import sys
from datetime import datetime, MAXYEAR
import time
import pytz
import pytest

from zhmcclient._utils import datetime_from_timestamp, \
    timestamp_from_datetime, datetime_to_isoformat, datetime_from_isoformat


# The Unix epoch
EPOCH_DT = datetime(1970, 1, 1, 0, 0, 0, 0, pytz.utc)

# HMC timestamp numbers (milliseconds since the Unix epoch)
TS_2000_02_29 = 951782400000  # last day in february in leap year
TS_2001_02_28 = 983318400000  # last day in february in non-leap year
TS_2017_08_15 = 1502755200000  # some day in summer
TS_2038_LIMIT = 2147483647000  # 2038-01-19 03:14:07 UTC (Year 2038 Problem)
TS_3000_12_31 = 32535129600000
TS_3001_LIMIT = 32535244799000  # 3001-01-01 07:59:59 UTC (Visual C++ limit)
TS_MAX = 253402300799999  # 9999-12-31 23:59:59.999 UTC

# Some durations in milliseconds
SEC_MS = 1000
MIN_MS = SEC_MS * 60
HOUR_MS = MIN_MS * 60
DAY_MS = HOUR_MS * 24

# Test cases for datetime / timestamp conversion functions
DATETIME_TIMESTAMP_TESTCASES = [
    # (datetime_tuple(y,m,d,h,m,s,us), timestamp)
    ((1970, 1, 1, 0, 0, 0, 0), 0),
    ((1970, 1, 1, 0, 0, 0, 123000), 123),
    ((1970, 1, 1, 0, 2, 3, 456000), 123456),
    ((1970, 1, 2, 10, 17, 36, 789000), 123456789),
    ((1973, 11, 29, 21, 33, 9, 123000), 123456789123),
    ((2000, 2, 29, 0, 0, 0, 0), TS_2000_02_29),
    ((2000, 3, 1, 0, 0, 0, 0), TS_2000_02_29 + DAY_MS),
    ((2001, 2, 28, 0, 0, 0, 0), TS_2001_02_28),
    ((2001, 3, 1, 0, 0, 0, 0), TS_2001_02_28 + DAY_MS),
    ((2017, 8, 15, 0, 0, 0, 0), TS_2017_08_15),
    ((2038, 1, 19, 3, 14, 7, 0), TS_2038_LIMIT),
    ((2038, 1, 19, 3, 14, 7, 1000), TS_2038_LIMIT + 1),
    ((3000, 12, 31, 23, 59, 59, 999000), TS_3000_12_31 + DAY_MS - 1),
    ((3001, 1, 1, 7, 59, 59, 0), TS_3001_LIMIT),
    ((3001, 1, 1, 8, 0, 0, 0), TS_3001_LIMIT + 1000),
    ((MAXYEAR - 1, 12, 31, 0, 0, 0, 0), TS_MAX + 1 - DAY_MS - 365 * DAY_MS),
    ((MAXYEAR, 12, 30, 0, 0, 0, 0), TS_MAX + 1 - 2 * DAY_MS),
    ((MAXYEAR, 12, 30, 23, 59, 59, 0),
     TS_MAX + 1 - 2 * DAY_MS + 23 * HOUR_MS + 59 * MIN_MS + 59 * SEC_MS),

    # The following testcases would be in range but are too close to the max
    # for pytz due to an implementation limitation: pytz.localize() checks the
    # input +/- 1 day.
    # ((MAXYEAR, 12, 31, 0, 0, 0, 0), TS_MAX + 1 - 1 * DAY_MS),
    # ((MAXYEAR, 12, 31, 23, 59, 59, 0), TS_MAX - 999),
    # ((MAXYEAR, 12, 31, 23, 59, 59, 998000), TS_MAX - 1),
    # ((MAXYEAR, 12, 31, 23, 59, 59, 999000), TS_MAX),
]


def find_max_value(test_func, initial_value):
    """
    Starting from an initial number (integer or float), find the maximum value
    for which the test function does not yet fail, and return that maximum
    value.
    """
    assert isinstance(initial_value, int) and initial_value > 0

    fails = FailsArray(test_func)
    value = initial_value

    # Advance the value exponentially beyond the max value
    while fails[value] == 0:
        value *= 2

    # Search for the exact max value in the previous range. We search for the
    # boundary where the fails array goes from 0 to 1.
    boundary = 0.5
    value = binary_search(fails, boundary, value // 2, value)
    max_value = value - 1

    # Verify that we found exactly the maximum:
    assert fails[max_value] == 0 and fails[max_value + 1] == 1, (
        f"max_value={max_value}, fails[+-2]: {fails[max_value - 2]}, "
        f"{fails[max_value - 1]}, {fails[max_value]}, "
        f"{fails[max_value + 1]}, {fails[max_value + 2]}")

    return max_value


class FailsArray:
    # pylint: disable=too-few-public-methods
    """
    An array that when accessed at an index returns 1 if the array value at
    that index causes an exception to be raised when passed to a test function,
    and 0 otherwise.
    """

    def __init__(self, test_func):
        """
        test_func is a test function that gets one integer argument and that
        raises an exception if the argument is out of range.
        """
        self.test_func = test_func

    def __getitem__(self, test_value):
        """
        Array access function, returning 1 if the given test value caused the
        test function to fail, and 0 otherwise.
        """
        try:
            self.test_func(test_value)
        except Exception:  # pylint: disable=broad-except
            return 1
        return 0


def binary_search(haystack, needle, lo, hi):
    """
    Binary search: Return index of value needle (or the next one if no exact
    match) in array haystack, within index range lo to hi.
    """
    while lo < hi:
        mid = (lo + hi) // 2
        if haystack[mid] > needle:
            hi = mid
        elif haystack[mid] < needle:
            lo = mid + 1
        else:
            return mid
    return hi


# TODO: Add testcases for repr_text() (currently indirectly tested)
# TODO: Add testcases for repr_list() (currently indirectly tested)
# TODO: Add testcases for repr_dict() (currently indirectly tested)
# TODO: Add testcases for repr_timestamp() (currently indirectly tested)
# TODO: Add testcases for repr_manager() (currently indirectly tested)


# Some tests for Python date & time related functions that we use.

def test_gmtime_epoch():
    """Test that time.gmtime() is based upon the UNIX epoch."""
    epoch_st = time.struct_time([1970, 1, 1, 0, 0, 0, 3, 1, 0])
    st = time.gmtime(0)
    assert st == epoch_st


def x_test_print_gmtime_max():
    """Print the maximum for time.gmtime()."""
    max_ts = find_max_value(time.gmtime, 1)
    max_st = time.gmtime(max_ts)
    print("\nMax Unix timestamp value for Python time.gmtime(): "
          f"{max_ts} ({max_st!r})")
    sys.stdout.flush()


def x_test_print_fromtimestamp_max():
    """Print the maximum for datetime.fromtimestamp(utc)."""

    def datetime_fromtimestamp_utc(ts):
        return datetime.fromtimestamp(ts, pytz.utc)

    max_ts = find_max_value(datetime_fromtimestamp_utc, 1)
    max_dt = datetime_fromtimestamp_utc(max_ts)
    print("\nMax Unix timestamp value for Python "
          f"datetime.fromtimestamp(utc): {max_ts} ({max_dt!r})")
    sys.stdout.flush()


def x_test_print_datetime_max():
    """Print datetime.max."""
    print(f"\nMax value for Python datetime (datetime.max): {datetime.max!r}")
    sys.stdout.flush()


# All tests for the datetime_from_timestamp() function.

@pytest.mark.parametrize(
    "tz_name", [None, 'UTC', 'US/Eastern', 'Europe/Berlin']
)
@pytest.mark.parametrize(
    "datetime_tuple, timestamp", DATETIME_TIMESTAMP_TESTCASES
)
def test_success_datetime_from_timestamp(
        datetime_tuple, timestamp, tz_name):
    """Test successful calls to datetime_from_timestamp()."""

    if os.name == 'nt' and timestamp > TS_3001_LIMIT:
        # Skip this test case, due to the lower limit on Windows
        return

    # Expected result, as timezone-unaware (but implied UTC)
    exp_dt_unaware = datetime(*datetime_tuple)

    # Expected result, as timezone-aware
    exp_dt = pytz.utc.localize(exp_dt_unaware)

    if tz_name is None:

        # Execute the code to be tested
        act_dt = datetime_from_timestamp(timestamp)

    else:
        tz = pytz.timezone(tz_name)

        # Execute the code to be tested
        act_dt = datetime_from_timestamp(timestamp, tz)

    # Verify the result.
    # Note: timezone-aware datetime objects compare equal according to
    # their effective point in time (e.g. as if normalized to UTC).
    assert act_dt == exp_dt


@pytest.mark.parametrize(
    "timestamp, exc_type", [
        (None, ValueError),
        (-1, ValueError),
        (TS_MAX + 1, ValueError),
    ]
)
def test_error_datetime_from_timestamp(timestamp, exc_type):
    """Test failing calls to datetime_from_timestamp()."""

    with pytest.raises(Exception) as exc_info:

        # Execute the code to be tested
        datetime_from_timestamp(timestamp)

    # Verify the result
    assert isinstance(exc_info.value, exc_type)


def x_test_print_max_datetime_from_timestamp():
    """Print the maximum for datetime_from_timestamp()."""
    max_ts = find_max_value(datetime_from_timestamp, 1)
    max_dt = datetime_from_timestamp(max_ts)
    print("\nMax HMC timestamp value for zhmcclient."
          f"datetime_from_timestamp(): {max_ts} ({max_dt!r})")
    sys.stdout.flush()


# All tests for the timestamp_from_datetime() function.

@pytest.mark.parametrize(
    "tz_name", [None, 'UTC', 'US/Eastern', 'Europe/Berlin']
)
@pytest.mark.parametrize(
    "datetime_tuple, timestamp", DATETIME_TIMESTAMP_TESTCASES
)
def test_success(datetime_tuple, timestamp, tz_name):
    """Test successful calls to timestamp_from_datetime()."""

    # Create a timezone-naive datetime object
    dt = datetime(*datetime_tuple)
    offset = 0  # because of default UTC

    if tz_name is not None:
        # Make the datetime object timezone-aware
        tz = pytz.timezone(tz_name)
        offset = tz.utcoffset(dt).total_seconds()
        dt = tz.localize(dt)

    exp_ts = timestamp - offset * 1000

    # Execute the code to be tested
    ts = timestamp_from_datetime(dt)

    # Verify the result
    assert ts == exp_ts


@pytest.mark.parametrize(
    "datetime_, exc_type", [
        (None, ValueError),
    ]
)
def test_error(datetime_, exc_type):
    """Test failing calls to timestamp_from_datetime()."""

    with pytest.raises(Exception) as exc_info:

        # Execute the code to be tested
        timestamp_from_datetime(datetime_)

    # Verify the result
    assert isinstance(exc_info.value, exc_type)


def test_datetime_max():
    """Test timestamp_from_datetime() with datetime.max."""

    # The test is that it does not raise an exception:
    timestamp_from_datetime(datetime.max)


# Test cases for datetime_to_isoformat()
TESTCASES_DATETIME_TO_ISOFORMAT = [
    # dt, exp_dt_str
    (
        datetime(2017, 9, 5, 12, 13, 10, 0),
        '2017-09-05 12:13:10'
    ),
    (
        datetime(2017, 9, 5, 12, 13, 10, 123456),
        '2017-09-05 12:13:10.123456'
    ),
    (
        datetime(2017, 9, 5, 12, 13, 10, 0, pytz.utc),
        '2017-09-05 12:13:10+00:00'
    ),
    (
        datetime(2017, 9, 5, 12, 13, 10, 0, pytz.timezone('CET')),
        '2017-09-05 12:13:10+01:00'
    ),
    (
        datetime(2017, 9, 5, 12, 13, 10, 0, pytz.timezone('EST')),
        '2017-09-05 12:13:10-05:00'
    ),
]


@pytest.mark.parametrize(
    "dt, exp_dt_str",
    TESTCASES_DATETIME_TO_ISOFORMAT)
def test_datetime_to_isoformat(dt, exp_dt_str):
    """
    Test function for datetime_to_isoformat().
    """

    # The function to be tested
    dt_str = datetime_to_isoformat(dt)

    assert dt_str == exp_dt_str


# Test cases for datetime_from_isoformat()
TESTCASES_DATETIME_FROM_ISOFORMAT = [
    # dt_str, exp_dt
    (
        '2017-09-05 12:13:10',
        datetime(2017, 9, 5, 12, 13, 10, 0)  # timezone-naive
    ),
    (
        '2017-09-05T12:13:10',
        datetime(2017, 9, 5, 12, 13, 10, 0)  # timezone-naive
    ),
    (
        '2017-09-05 12:13:10.123456',
        datetime(2017, 9, 5, 12, 13, 10, 123456)  # timezone-naive
    ),
    (
        '2017-09-05 12:13:10+00:00',
        datetime(2017, 9, 5, 12, 13, 10, 0, pytz.utc)
    ),
    (
        '2017-09-05 12:13:10+0000',
        datetime(2017, 9, 5, 12, 13, 10, 0, pytz.utc)
    ),
    (
        '2017-09-05 12:13:10+01:00',
        datetime(2017, 9, 5, 12, 13, 10, 0, pytz.timezone('CET'))
    ),
    (
        '2017-09-05 12:13:10+0100',
        datetime(2017, 9, 5, 12, 13, 10, 0, pytz.timezone('CET'))
    ),
    (
        '2017-09-05 12:13:10-05:00',
        datetime(2017, 9, 5, 12, 13, 10, 0, pytz.timezone('EST'))
    ),
]


@pytest.mark.parametrize(
    "dt_str, exp_dt",
    TESTCASES_DATETIME_FROM_ISOFORMAT)
def test_datetime_from_isoformat(dt_str, exp_dt):
    """
    Test function for datetime_from_isoformat().
    """

    # The function to be tested
    dt = datetime_from_isoformat(dt_str)

    assert dt == exp_dt
